{
 BUSINESS CONSULTING
 s a i n t - p e t e r s b u r g

         Components Library for Borland Delphi 4.x, 5.x
         Copyright (c) 1998-2000 Alex'EM

}
unit DCFileCtrl;

interface

uses
  Windows, Classes, SysUtils, Controls, Forms, ShellAPI, ShlObj, CommCtrl,
  ComCtrls, ActiveX, ComObj, Graphics, DCConst, DCChoice, DCPopupWindow, Messages;

const
  BIF_USENEWUI = $0040;

type
  PFolderItem_tag = ^TFolderItem;
  TFolderItem = record
     SPIDL: PItemIDList;
     FPIDL: PItemIDList;
     DisplayName: string;
     DisplayPath: string;
     Attrs: ULONG;
  end;

  TDCBrowseFolder = class(TDCCustomTreeEdit)
  private
    FIDesktopFolder: IShellFolder;
    FImages: TImageList;
    FBitmap: TBitmap;
    procedure LoadIDList(Node: TTreeNode);
    function GetTreeView: TTreeView;
    procedure DoCreate(Sender: TObject);
    procedure DoCancel(Sender: TObject);
    procedure DrawButton(Sender: TCustomTreeView; ARect: TRect; Node: TTreeNode);
    procedure DrawImage(Sender: TCustomTreeView;  NodeRect: TRect; ImageIndex: Integer;
      Attrs: ULONG);
    function ClearNode(Node: TTreeNode; FreeNodeData: boolean): TTreeNode;
  protected
    procedure DefineBtnChoiceStyle; override;
    procedure SetText(Value: string); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Expanding(Sender: TObject; Node: TTreeNode; var AllowExpansion: Boolean); override;
    procedure GetHintOnError; override;
    procedure DoShowError(AErrorWindow: TDCMessageWindow); override;
    procedure CMSetSelection(var Message: TMessage); message CM_SETSELECTION;
    procedure CustomDrawItem(Sender: TCustomTreeView; Node: TTreeNode;
       State: TCustomDrawState; var DefaultDraw: Boolean);  override;
    procedure ClearTreeItems; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetNode(Value: string; var Node: TTreeNode; var ErrorCode: integer): boolean; override;
    function SelectFolder(Name: string): boolean;
    procedure InitTree; override;
    property TreeView: TTreeView read GetTreeView;
    property Images;
  published
    property DrawStyle;  
  end;

function CreateDirectoryEx(FilePath: string; SecurityAttributes: PSECURITYATTRIBUTES = nil): boolean;
function SelectDirectoryEx(const Caption: string; const Root: WideString;
  var Directory: string): Boolean;

implementation

uses DCEditTools;

var
 AIShellFolder: IShellFolder;

function DirectoryExists(const Name: string): Boolean;
 var
  AIDesktopFolder: IShellFolder;
  WideValue: WideString;
  Eaten, Flags: LongWord;
  ItemIDList: PItemIDList;
begin
  OLECheck(SHGetDesktopFolder(AIDesktopFolder));
  if Name <> '' then
  begin
    WideValue := Name;
    Result := (AIDesktopFolder.ParseDisplayName(Application.Handle, nil, POleStr(WideValue),
      Eaten, ItemIDList, Flags) = NOERROR);
    if Result then
    begin
      Flags := $FFFFFFFF;
      AIDesktopFolder.GetAttributesOf(1, ItemIDList, Flags);
      Result := (Flags and SFGAO_FOLDER) = SFGAO_FOLDER;
    end;
  end
  else
    Result := False;
end;

function CreateDirectoryEx(FilePath: string; SecurityAttributes: PSECURITYATTRIBUTES = nil): boolean;
begin
  if Length(FilePath) = 0 then
  begin
    Result := False;
    Exit;
  end;

  if (AnsiLastChar(FilePath) <> nil) and (AnsiLastChar(FilePath)^ = '\') then
    Delete(FilePath, Length(FilePath), 1);

  if (Length(FilePath) < 3) or DirectoryExists(FilePath) or
     (ExtractFilePath(FilePath) = FilePath) then
  begin
    Result := True;
    Exit;
  end;

  Result := CreateDirectoryEx(ExtractFilePath(FilePath), SecurityAttributes);
  if Result and not CreateDirectory(PChar(FilePath), SecurityAttributes) then
    Result:= False;
end;

{ TDCDirectoryEdit }
function NextPIDL(IDList: PItemIDList): PItemIDList;
begin
  Result := IDList;
  Inc(PChar(Result), IDList^.mkid.cb);
end;

function GetPIDLSize(IDList: PItemIDList): Integer;
begin
  Result := 0;
  if Assigned(IDList) then
  begin
    Result := SizeOf(IDList^.mkid.cb);
    while IDList^.mkid.cb <> 0 do
    begin
      Result := Result + IDList^.mkid.cb;
      IDList := NextPIDL(IDList);
    end;
  end;
end;

function CreatePIDL(Size: Integer): PItemIDList;
var
  Malloc: IMalloc;
  HR: HResult;
begin
  Result := nil;

  HR := SHGetMalloc(Malloc);
  if Failed(HR) then Exit;

  try
    Result := Malloc.Alloc(Size);
    if Assigned(Result) then
      FillChar(Result^, Size, 0);
  finally
  end;
end;

procedure FreePIDL(PIDL: PItemIDList);
var
  Malloc: IMalloc;
  HR: HResult;
begin
  HR := SHGetMalloc(Malloc);
  if Failed(HR) then Exit;
  try
    Malloc.Free(PIDL);
  finally
  end;
end;

function ConcatPIDL(IDList1, IDList2: PItemIDList): PItemIDList;
var
  cb1, cb2: Integer;
begin
  if Assigned(IDList1) then
    cb1 := GetPIDLSize(IDList1) - SizeOf(IDList1^.mkid.cb)
  else
    cb1 := 0;

  cb2 := GetPIDLSize(IDList2);

  Result := CreatePIDL(cb1 + cb2);
  if Assigned(Result) then
  begin
    if Assigned(IDList1) then
      CopyMemory(Result, IDList1, cb1);
    CopyMemory(PChar(Result) + cb1, IDList2, cb2);
  end;
end;


function GetDisplayName(ShellFolder: IShellFolder; PIDL: PItemIDList;
  lType: integer = SHGDN_NORMAL): string;
 var
  StrRet: TStrRet;
  P: PChar;
begin
  Result := '';
  ShellFolder.GetDisplayNameOf(PIDL, lType, StrRet);
  case StrRet.uType of
    STRRET_CSTR:
      SetString(Result, StrRet.cStr, lStrLen(StrRet.cStr));
    STRRET_OFFSET:
      begin
        P := @PIDL.mkid.abID[StrRet.uOffset - SizeOf(PIDL.mkid.cb)];
        Result := P;
        //SetString(Result, P, PIDL.mkid.cb - StrRet.uOffset);
      end;
    STRRET_WSTR:
      Result := StrRet.pOleStr;
  end;
end;

function GetImageIndex(PIDL: PItemIDList): integer;
 var
  FileInfo: TShFileInfo;
begin
  SHGetFileInfo(PChar(PIDL), 0, FileInfo, SizeOf(TSHFileInfo),
    SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
  Result := FileInfo.iIcon;
end;

function GetSelectedIndex(PIDL: PItemIDList): integer;
 var
  FileInfo: TShFileInfo;
begin
  SHGetFileInfo(PChar(PIDL), 0, FileInfo, SizeOf(TSHFileInfo),
    SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON or SHGFI_OPENICON);
  Result := FileInfo.iIcon;
end;

function CustomSortProc(Node1, Node2: TTreeNode; Data: Longint): integer; stdcall;

begin
  Result := SmallInt(AIShellFolder.CompareIDs(
                  0,
                  PFolderItem_tag(Node1.Data).SPIDL,
                  PFolderItem_tag(Node2.Data).SPIDL));
end;

procedure TDCBrowseFolder.ClearTreeItems;
 var
  Node: TTreeNode;
begin
  with TreeView do
  begin
    if Items.Count > 0 then
    begin
      Node := Items.GetFirstNode;
      while Node <> nil do Node := ClearNode(Node, True);
    end;
  end;
  inherited;
end;

procedure TDCBrowseFolder.CMSetSelection(var Message: TMessage);

  function LocatePIDL(SIDList: PItemIDList; PNode: TTreeNode): boolean;
   var
    Node: TTreeNode;
    FolderItem: PFolderItem_tag;
  begin
    Result := False;
    if PNode = nil then Exit;
    if PNode.HasChildren and (PNode.GetFirstChild = nil) then
    begin
      PNode.HasChildren := False;
      LoadIDList(PNode);
    end;
    Node := PNode.GetFirstChild;
    while (Node <> nil) do
    begin
      FolderItem := Node.Data;
      if FIDesktopFolder.CompareIDs(0, FolderItem.SPIDL, SIDList) = 0 then
      begin
        TreeView.Selected := Node;
        Result := True;
        Break;
      end
      else begin
        {       PIDL'}
      end;
      Node := Node.GetNextSibling;
    end;
  end;

begin
  if not TreeInitialized then InitTree;
  LocatePIDL(PItemIDList(Message.LParam), TreeView.Items.GetFirstNode);
end;

constructor TDCBrowseFolder.Create(AOwner: TComponent);
begin
  inherited;
  Style := teDropDown;
  OLECheck(SHGetDesktopFolder(FIDesktopFolder));
  AIShellFolder := FIDesktopFolder;
  FImages := TImageList.Create(nil);
  TreeView.ShowRoot := False;
  DBObject.Caption := LoadStr(RES_BFLD_DBOCAPTION);
  FBitmap := TBitmap.Create;
end;

procedure TDCBrowseFolder.CustomDrawItem(Sender: TCustomTreeView; Node: TTreeNode;
       State: TCustomDrawState; var DefaultDraw: Boolean);
 var
  PFolderItem: PFolderItem_tag;
  NodeRect: TRect;
begin
  inherited;
  if DefaultDraw then begin
    DefaultDraw := False;
    PFolderItem := Node.Data;
    with PFolderItem^,  Sender.Canvas do
    begin
      Brush.Color := TreeView.Color;

      Font.Assign(TreeView.Font);
      if Attrs and SFGAO_COMPRESSED <> 0 then
      begin
        Font.Color := clBlue;
      end;
      if cdsSelected in State then Font.Color := clCaptionText;

      if cdsSelected in State then
        Brush.Color := clHighlight
      else
        Brush.Color := TreeView.Color;
      NodeRect := Node.DisplayRect(True);

      if Node.Level = 0 then
        OffsetRect(NodeRect, 6, 0)
      else
        OffsetRect(NodeRect, 4, 0);
      FillRect(NodeRect);

      InflateRect(NodeRect, -2, -1);
      TextOut(NodeRect.Left, NodeRect.Top, Node.Text);
      InflateRect(NodeRect, 2, 1);

      if cdsFocused in State then DrawFocusRect(NodeRect);

      NodeRect := Node.DisplayRect(False);

      if Node.Level > 0 then
      begin
        NodeRect.Left := NodeRect.Left + ((Node.Level -1) * TreeView.Indent + 5);
        DrawButton(Sender, NodeRect, Node);
        NodeRect.Left := NodeRect.Left + TreeView.Indent;
      end
      else
        NodeRect.Left := NodeRect.Left + 7;

      if cdsSelected in State then
        DrawImage(Sender, NodeRect, Node.SelectedIndex, Attrs)
      else
        DrawImage(Sender, NodeRect, Node.ImageIndex, Attrs);
    end;
  end;
end;

procedure TDCBrowseFolder.DefineBtnChoiceStyle;
begin
  if BtnChoiceAssigned then
  begin
    Glyph.LoadFromResourceName(HInstance, 'DC_BTNFOLDER');
    ButtonStyle := esDropDown;
    ButtonChoiceStyle := btsCustom;
  end;
end;

destructor TDCBrowseFolder.Destroy;
begin
  FImages.Handle := 0;
  FImages.Free;
  FBitmap.Free;
  inherited;
end;

procedure TDCBrowseFolder.DoCancel(Sender: TObject);
begin
  HideErrorMessage;
end;

procedure TDCBrowseFolder.DoCreate(Sender: TObject);
begin
  if not CreateDirectoryEx(PChar(Text), nil) then
  begin
    HideErrorMessage;
    ErrorCode := ERR_BFLD_NOTFOLDER;
    ShowErrorMessage;
  end
  else
    HideErrorMessage;
end;

procedure TDCBrowseFolder.DoShowError(AErrorWindow: TDCMessageWindow);
begin
  if ErrorCode = ERR_TREE_ILLIGALVALUE then
  begin
    AErrorWindow.AddButton('#Create', '', LoadStr(RES_STRN_VAL_CREATE), DoCreate);
    AErrorWindow.AddButton('#Cancel', '', LoadStr(RES_STRN_VAL_CANCEL), DoCancel);
  end;
  inherited;
end;

procedure TDCBrowseFolder.DrawButton(Sender: TCustomTreeView; ARect: TRect; Node: TTreeNode);
 var
  ButtonSize, cx, cy, wx, hy: Integer;

  procedure DrawDotedLine(Canvas: TCanvas; AColor: TColor; ABegin: TPoint;
    ALength, Direction: integer);
   var
    i: integer;
  begin
    case Direction of
      0:
        for i := 0 to ALength do
          if i mod 2 = 0 then
            Canvas.Pixels[ABegin.X + i, ABegin.Y] := AColor;
      1:
        for i := 0 to ALength do
          if i mod 2 = 0 then
            Canvas.Pixels[ABegin.X, ABegin.Y + i] := AColor;
    end;
  end;
begin
  ButtonSize := 5;

  wx := ARect.Left + TreeView.Indent + ButtonSize;
  hy := ARect.Bottom;
  cx := ARect.Left + TreeView.Indent div 2;
  cy := ARect.Top + (ARect.Bottom - ARect.Top) div 2;

  with Sender.Canvas do
  begin
    Pen.Style := psSolid;

    if Node.HasChildren then
      DrawDotedLine(Sender.Canvas, clAppWorkSpace, Point(cx + 5, cy), wx - (cx + 5), 0)
    else
      DrawDotedLine(Sender.Canvas, clAppWorkSpace, Point(cx, cy), wx - cx, 0);

    DrawDotedLine(Sender.Canvas, clAppWorkSpace, Point(cx, ARect.Top), hy - cy - ButtonSize, 1);

    if ((Node.GetNextVisible <> nil) and (Node.GetNextVisible.Level = Node.Level)) or
       (Node.GetNextSibling <> nil) then
      DrawDotedLine(Sender.Canvas, clAppWorkSpace, Point(cx, cy + ButtonSize-1), hy - cy - ButtonSize, 1);

    if Node.HasChildren then
    begin
      Pen.Color := clAppWorkSpace;

      MoveTo(cx-ButtonSize+1, cy-ButtonSize+1);

      LineTo(cx-ButtonSize+1, cy+ButtonSize-1);
      LineTo(cx+ButtonSize-1, cy+ButtonSize-1);
      LineTo(cx+ButtonSize-1, cy-ButtonSize+1);
      LineTo(cx-ButtonSize+1, cy-ButtonSize+1);

      Pen.Color := clBlack;
      PenPos := Point(cx-ButtonSize+3, cy);
      LineTo(cx+ButtonSize-2, cy);
      if not Node.Expanded then
      begin
        PenPos := Point(cx, cy-ButtonSize+3);
        LineTo(cx, cy+ButtonSize-2);
      end
      else begin
        Pen.Color := TreeView.Color;
        PenPos := Point(cx, cy-ButtonSize+3);
        LineTo(cx, cy);
        PenPos := Point(cx, cy + 1);
        LineTo(cx, cy+ButtonSize-2);
      end;
    end
    else begin
      if Node.GetNextSibling <> nil then
        DrawDotedLine(Sender.Canvas, clAppWorkSpace, Point(cx, cy-ButtonSize+1), ButtonSize+2, 1)
      else
        DrawDotedLine(Sender.Canvas, clAppWorkSpace, Point(cx, cy-ButtonSize+1), (ButtonSize+2) div 2, 1);
    end;

    Node := Node.Parent;
    while Node <> nil do
    begin
      cx := cx - TreeView.Indent;
      if Node.GetNextSibling <> nil then
        DrawDotedLine(Sender.Canvas, clAppWorkSpace, Point(cx, ARect.Top), hy - ARect.Top, 1);
      Node := Node.Parent;
    end;

  end;
end;

procedure TDCBrowseFolder.DrawImage(Sender: TCustomTreeView; NodeRect: TRect;
  ImageIndex: Integer; Attrs: ULONG);
var
  cy: Integer;
begin
  cy := NodeRect.Top + (NodeRect.Bottom - NodeRect.Top) div 2;

  FBitmap.Width  := Images.Width;
  FBitmap.Height := Images.Height;

  try
    if Attrs and SFGAO_SHARE <> 0 then
      Images.DrawOverlay(FBitmap.Canvas, 0, 0, ImageIndex, 0)
    else
      Images.Draw(FBitmap.Canvas, 0, 0, ImageIndex, True);

    if Attrs and SFGAO_HIDDEN <> 0 then
      TransformBitmapTransparent(FBitmap, nil, FBitmap, 60);

    Sender.Canvas.Draw(NodeRect.Left, cy - Images.Height div 2, FBitmap);
  finally
    {}
  end;
end;

procedure TDCBrowseFolder.Expanding(Sender: TObject; Node: TTreeNode;
  var AllowExpansion: Boolean);
begin
  with TreeView do
  begin
    Items.BeginUpdate;
    try
      if Node.HasChildren and (Node.GetFirstChild = nil) then
      begin
        Node.HasChildren := False;
        LoadIDList(Node);
      end;
    finally
      Items.EndUpdate;
    end;
  end;
  inherited;
end;

function TDCBrowseFolder.ClearNode(Node: TTreeNode; FreeNodeData: boolean): TTreeNode;
 var
  FolderItem: PFolderItem_tag;
  ChildNode: TTreeNode;
begin
  if Node <> nil then
  begin
    {   Child}
    if Node.HasChildren then
    begin
      ChildNode := Node.GetFirstChild;
      while ChildNode <> nil do ChildNode := ClearNode(ChildNode, True);
    end;
    {   }
    if FreeNodeData then
    begin
      FolderItem := Node.Data;
      FreePIDL(FolderItem.SPIDL);
      FreePIDL(FolderItem.FPIDL);
      FreeMem(FolderItem);
      Node.Data := nil;
    end;
    Result := Node.GetNextSibling;
  end
  else
    Result := nil;
end;

procedure TDCBrowseFolder.GetHintOnError;
begin
  case ErrorCode of
    ERR_TREE_ILLIGALVALUE:
      begin
        ErrorHint := Format(LoadStr(RES_FOLD_ERR_WRONG), [Text]);
        Exit;
      end;
    ERR_BFLD_NOTFOLDER:
      begin
        ErrorHint := Format(LoadStr(RES_FOLD_ERR_NOTFLD), [Text]);
        Exit;
      end;
  else
    ErrorHint := '';
  end;
  inherited;
end;

function TDCBrowseFolder.GetNode(Value: string;
  var Node: TTreeNode; var ErrorCode: integer): boolean;
 var
  WideValue: WideString;
  Eaten, Flags: LongWord;
  ItemIDList: PItemIDList;
begin
  Node   := nil;
  Result := True;

  if Value <> '' then
  begin
    WideValue := Value;
    Result := (FIDesktopFolder.ParseDisplayName(Application.Handle, nil, POleStr(WideValue),
      Eaten, ItemIDList, Flags) = NOERROR);
    if Result then
    begin
      FIDesktopFolder.GetAttributesOf(1, ItemIDList, Flags);
      Result := (Flags and SFGAO_FOLDER) = SFGAO_FOLDER;
      if not Result then ErrorCode := ERR_BFLD_NOTFOLDER;
    end;
  end;
end;

function TDCBrowseFolder.GetTreeView: TTreeView;
begin
  Result := inherited GetTreeView;
end;

procedure TDCBrowseFolder.InitTree;
 var
  RootPIDL: PItemIDList;
  DesktopNode, FolderNode, NetworkNode: TTreenode;
  PFolderItem: PFolderItem_tag;
  Attrs: ULONG;
begin
  with TreeView do
  begin
    Items.BeginUpdate;

    {Desktop}
    OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_DESKTOP, RootPIDL));
    PFolderItem := New(PFolderItem_tag);
    PFolderItem.SPIDL := RootPIDL;
    PFolderItem.FPIDL := ConcatPIDL(nil, RootPIDL);
    PFolderItem.DisplayName := GetDisplayName(AIShellFolder, RootPIDL);
    PFolderItem.DisplayPath := '';
    AIShellFolder.GetAttributesOf(1, RootPIDL, Attrs);
    PFolderItem.Attrs := Attrs;

    DesktopNode := Items.AddFirst(nil, PFolderItem.DisplayName);
    DesktopNode.ImageIndex    := GetImageIndex(RootPIDL);
    DesktopNode.SelectedIndex := DesktopNode.ImageIndex;
    DesktopNode.Data := PFolderItem;

    {Drivers}
    OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_DRIVES, RootPIDL));
    PFolderItem := New(PFolderItem_tag);
    PFolderItem.SPIDL := RootPIDL;
    PFolderItem.FPIDL := ConcatPIDL(nil, RootPIDL);
    PFolderItem.DisplayName := GetDisplayName(AIShellFolder, RootPIDL);
    PFolderItem.DisplayPath := '';
    AIShellFolder.GetAttributesOf(1, RootPIDL, Attrs);
    PFolderItem.Attrs := Attrs;

    FolderNode := Items.AddChild(DesktopNode, PFolderItem.DisplayName);
    FolderNode.ImageIndex    := GetImageIndex(RootPIDL);
    FolderNode.SelectedIndex := FolderNode.ImageIndex;
    FolderNode.HasChildren   := True;
    FolderNode.Data := PFolderItem;

    {Network}
    OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_NETWORK, RootPIDL));
    PFolderItem := New(PFolderItem_tag);
    PFolderItem.SPIDL := RootPIDL;
    PFolderItem.FPIDL := ConcatPIDL(nil, RootPIDL);
    PFolderItem.DisplayName := GetDisplayName(AIShellFolder, RootPIDL);
    PFolderItem.DisplayPath := '';
    AIShellFolder.GetAttributesOf(1, RootPIDL, Attrs);
    PFolderItem.Attrs := Attrs;

    NetworkNode := Items.AddChild(DesktopNode, PFolderItem.DisplayName);
    NetworkNode.ImageIndex    := GetImageIndex(RootPIDL);
    NetworkNode.SelectedIndex := NetworkNode.ImageIndex;
    NetworkNode.Data := PFolderItem;
    NetworkNode.HasChildren   := True;

    DesktopNode.Expand(False);
    FolderNode.Expand(False);

    TreeView.Selected := FolderNode;
    TreeView.TopItem  := DesktopNode;
    Items.EndUpdate;
  end;
  inherited;
end;

procedure TDCBrowseFolder.LoadIDList(Node: TTreeNode);
 var
  SaveCursor: TCursor;
  EnumList: IEnumIDList;
  PIDL: PItemIDList;
  PFolderItem: PFolderItem_tag;
  NumIDCount: DWORD;
  ChildNode: TTreeNode;
  Attrs: Ulong;
begin
  SaveCursor := Screen.Cursor;
  try
    if Node.HasChildren then ClearNode(Node, False);
    Screen.Cursor := crHourglass;

    PIDL := PFolderItem_tag(Node.Data).FPIDL;

    if FIDesktopFolder.BindToObject(PIDL, nil, IID_IShellFolder, Pointer(AIShellFolder)) <> S_OK then
    begin
      Screen.Cursor := SaveCursor;
      Exit;
    end;
    try
      OLECheck(AIShellFolder.EnumObjects(Application.Handle,
        SHCONTF_FOLDERS + SHCONTF_INCLUDEHIDDEN, EnumList));
    except
      Exit;
    end;
    with TreeView do
    begin
      while EnumList.Next(1, PIDL, NumIDCount) = S_OK do
      begin
        Attrs := $FFFFFFFF;
        AIShellFolder.GetAttributesOf(1, PIDL, Attrs);
        if Attrs and SFGAO_HASPROPSHEET <> 0 then
        begin
          PFolderItem := New(PFolderItem_tag);
          PFolderItem.SPIDL := PIDL;
          PFolderItem.FPIDL := ConcatPIDL(PFolderItem_tag(Node.Data).FPIDL, PIDL);
          PFolderItem.DisplayName := GetDisplayName(AIShellFolder, PIDL);
          PFolderItem.DisplayPath := GetDisplayName(AIShellFolder, PIDL, SHGDN_FORPARSING);
          PFolderItem.Attrs       := Attrs;

          ChildNode := Items.AddChild(Node, PFolderItem.DisplayName);
          ChildNode.Data := PFolderItem;
          ChildNode.ImageIndex := GetImageIndex(PFolderItem.FPIDL);
          ChildNode.SelectedIndex := GetSelectedIndex(PFolderItem.FPIDL);
          ChildNode.HasChildren := Attrs and SFGAO_HASSUBFOLDER <> 0;
        end;
      end;
    end;
  finally
    Node.CustomSort(@CustomSortProc, 0);
    Screen.Cursor := SaveCursor;
  end;
end;

function TDCBrowseFolder.SelectFolder(Name: string): boolean;
 var
  WideValue: WideString;
  Eaten, Flags: LongWord;
  ItemIDList: PItemIDList;
begin
  {sorry. Not supported yet}
  Result := True;
  if Name <> '' then
  begin
    WideValue := Name;
    Result := (FIDesktopFolder.ParseDisplayName(Application.Handle, nil, POleStr(WideValue),
      Eaten, ItemIDList, Flags) = NOERROR);
    if Result then
    begin
      FIDesktopFolder.GetAttributesOf(1, ItemIDList, Flags);
      Result := (Flags and SFGAO_FOLDER) = SFGAO_FOLDER;
      if Result then
      begin
        SendMessage(Handle, CM_SETSELECTION, 0, LPARAM(ItemIDList));
      end;
    end;
  end;
end;

procedure TDCBrowseFolder.SetParent(AParent: TWinControl);
 var
  FileInfo: TShFileInfo;
begin
  inherited;
  if AParent <> nil then
  begin
    FImages.Handle := SHGetFileInfo('C:\', 0, FileInfo, SizeOf(FileInfo),
      SHGFI_SYSICONINDEX or SHGFI_SMALLICON);

    FImages.ShareImages := True;
    FImages.BlendColor := clHighLight;

    Images := FImages;
  end;
end;

procedure TDCBrowseFolder.SetText(Value: string);
begin
  if TreeView.Selected <> nil then
    Value := PFolderItem_tag(TreeView.Selected.Data).DisplayPath;
  inherited SetText(Value);
end;

procedure BFFMInitialized(Wnd: HWnd; lData: LPARAM);
 var
  Rect: TRect;
  IDesktopFolder: IShellFolder;
  Eaten, Flags: LongWord;
  SelectionPIDL: PItemIDList;
begin
  GetWindowRect(Wnd, Rect);
  SetWindowPos(Wnd, 0,
    (GetSystemMetrics(SM_CXSCREEN) - Rect.Right + Rect.Left) div 2,
    (GetSystemMetrics(SM_CYSCREEN) - Rect.Bottom + Rect.Top) div 2,
    0, 0, SWP_NOACTIVATE or SWP_NOSIZE or SWP_NOZORDER);

  if POleStr(lData) <> '' then
  begin
    SHGetDesktopFolder(IDesktopFolder);
    if IDesktopFolder.ParseDisplayName(Wnd, nil, POleStr(lData), Eaten,
      SelectionPIDL, Flags) = NO_ERROR then
    begin
      SendMessage(Wnd, BFFM_SETSELECTION, 0, LPARAM(SelectionPIDL));
    end;
  end;
end;

function BrowseCallback(Wnd: HWnd; Msg: UINT; lParam: LPARAM; lData: LPARAM): integer; stdcall;
begin
  Result := 0;
  case Msg of
    BFFM_INITIALIZED: BFFMInitialized(Wnd, lData);
  end;
end;

function SelectDirectoryEx(const Caption: string; const Root: WideString;
  var Directory: string): Boolean;
var
  WindowList: Pointer;
  BrowseInfo: TBrowseInfo;
  Buffer: PChar;
  IDesktopFolder: IShellFolder;
  Eaten, Flags: LongWord;
  RootItemIDList, ItemIDList: PItemIDList;
  ShellMalloc: IMalloc;
  BrowseDirectory: WideString;
begin
  Result := False;
  FillChar(BrowseInfo, SizeOf(BrowseInfo), 0);
  if (ShGetMalloc(ShellMalloc) = S_OK) and (ShellMalloc <> nil) then
  begin
    Buffer := ShellMalloc.Alloc(MAX_PATH);
    try
      RootItemIDList := nil;
      if Root <> '' then
      begin
        SHGetDesktopFolder(IDesktopFolder);
        IDesktopFolder.ParseDisplayName(Application.Handle, nil,
          POleStr(Root), Eaten, RootItemIDList, Flags);
      end;
      BrowseDirectory := Directory;
      with BrowseInfo do
      begin
        hwndOwner := Application.Handle;
        pidlRoot := RootItemIDList;
        pszDisplayName := Buffer;
        lpszTitle := PChar(Caption);
        ulFlags := BIF_RETURNONLYFSDIRS or BIF_EDITBOX or BIF_USENEWUI;
        lpfn := BrowseCallback;
        lParam := Integer(Pointer(BrowseDirectory));
      end;
      WindowList := DisableTaskWindows(0);
      try
        OleCheck(CoInitialize(nil));
        ItemIDList := ShBrowseForFolder(BrowseInfo);
      finally
        EnableTaskWindows(WindowList);
      end;
      Result :=  ItemIDList <> nil;
      if Result then
      begin
        ShGetPathFromIDList(ItemIDList, Buffer);
        ShellMalloc.Free(ItemIDList);
        Directory := Buffer;
      end;
    finally
      ShellMalloc.Free(Buffer);
    end;
  end;
end;

end.
